# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import types
from functools import partial
from typing import Any, Dict, List, Optional

import torch
import torch.nn as nn
from torch.distributed.fsdp import CPUOffload, FullyShardedDataParallel, MixedPrecision, ShardingStrategy
from torch.distributed.fsdp._common_utils import _get_module_fsdp_state_if_fully_sharded_module
from torch.distributed.fsdp._runtime_utils import _lazy_init
from torch.distributed.fsdp.wrap import lambda_auto_wrap_policy
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.checkpoint import create_selective_checkpoint_contexts, noop_context_fn

from ..models import load_model_weights
from ..utils import logging
from ..utils.import_utils import is_torch_version_greater_than
from .checkpoint import CheckpointFunction
from .fsdp import (
    clip_grad_norm_,
    init_fsdp_fn,
    parallel_init_fsdp_fn,
    parallel_load_safetensors,
    register_checkpoint_extension,
)
from .parallel_state import get_parallel_state
from .utils import get_module_from_path, set_module_from_path


if is_torch_version_greater_than("2.4"):
    from torch.distributed._composable.fsdp import MixedPrecisionPolicy, fully_shard
    from torch.distributed.tensor.parallel import parallelize_module


logger = logging.get_logger(__name__)


def verbose_fsdp_grouping(model, prefix="", depth=0):
    indent = "    " * depth

    for name, child in model.named_children():
        if isinstance(child, FullyShardedDataParallel):
            module_names = [m_name for m_name, _ in child.named_modules()][1:]  # [1:] 排除自身
            strategy = child.sharding_strategy
            logger.debug_rank0(f"{indent}├── [FSDP Group] {prefix}{name}")
            logger.debug_rank0(
                f"{indent}│   ├── Sharding Strategy: {strategy}, Mixed Precision: {child.mixed_precision}"
            )
            logger.debug_rank0(f"{indent}│   └── Contains Modules: {module_names}")

            verbose_fsdp_grouping(child, prefix=f"{prefix}{name}.", depth=depth + 1)
        else:
            verbose_fsdp_grouping(child, prefix=f"{prefix}{name}.", depth=depth)


def build_parallelize_model(
    model: "nn.Module",
    weights_path: Optional[str] = None,
    sharding_plan: Optional[Dict[str, Any]] = None,
    enable_full_shard: bool = True,
    enable_mixed_precision: bool = True,
    enable_gradient_checkpointing: bool = True,
    basic_modules: Optional[List[str]] = None,
    **kwargs,
) -> "nn.Module":
    """
    Applies parallel strategies to the model.
    """
    parallel_state = get_parallel_state()
    fsdp_no_shard_states = None

    if not parallel_state.fsdp_enabled:
        if kwargs.get("init_device") != "cuda":
            raise ValueError("Only FSDP training supports `init_device=cpu` or `init_device=meta`.")
        if kwargs.pop("enable_fsdp_offload", False):
            raise ValueError("Only FSDP training supports `enable_fsdp_offload`.")

    if enable_mixed_precision:  # upcast to float32 before feed it to optimizer
        model = model.float()

    if enable_gradient_checkpointing and hasattr(model, "gradient_checkpointing_enable"):
        logger.info_rank0("Enable gradient checkpointing.")
        use_reentrant = kwargs.pop("enable_reentrant", False)
        if use_reentrant:
            torch.utils.checkpoint.CheckpointFunction = CheckpointFunction

        ops_to_save = kwargs.pop("ops_to_save", None)
        context_fn = (
            partial(create_selective_checkpoint_contexts, ops_to_save) if ops_to_save is not None else noop_context_fn
        )
        model.gradient_checkpointing_enable(
            gradient_checkpointing_kwargs={"use_reentrant": use_reentrant, "context_fn": context_fn}
        )

    if parallel_state.tp_enabled:
        logger.info_rank0("Apply tensor parallel to the model.")
        model = parallelize_module(
            model,
            device_mesh=parallel_state.tp_mesh,
        )

    if parallel_state.ep_enabled:
        parallel_plan = model.get_parallel_plan()
        ep_param_suffix = parallel_plan.ep_param_suffix

        fqn2spec_info = parallel_plan.apply(model, parallel_state.ep_fsdp_device_mesh)
        fsdp_no_shard_states_fqn_to_module = parallel_plan.get_fsdp_no_shard_info(model)

        fsdp_no_shard_states = list(fsdp_no_shard_states_fqn_to_module.values())
        fsdp_no_shard_states_fqn = list(fsdp_no_shard_states_fqn_to_module.keys())
        logger.info_rank0(f"Apply expert parallel to the model successfully.\nEP modules: {fsdp_no_shard_states_fqn}.")
    else:
        fqn2spec_info = None
        ep_param_suffix = None
        fsdp_no_shard_states = None
        fsdp_no_shard_states_fqn = None

    if parallel_state.fsdp_enabled:
        logger.info_rank0(f"Apply data parallel to the model: {parallel_state.dp_mode}.")
        if parallel_state.dp_mode == "fsdp2":
            fsdp_kwargs = {
                "mesh": parallel_state.fsdp_mesh,
                "reshard_after_forward": enable_full_shard,
                **kwargs.pop("fsdp_kwargs", {}),
            }
            if enable_mixed_precision:
                logger.info_rank0("Enable mixed precision training.")
                mp_policy = MixedPrecisionPolicy(
                    param_dtype=torch.bfloat16,
                    reduce_dtype=torch.float32,
                    output_dtype=torch.bfloat16,
                )
                fsdp_kwargs["mp_policy"] = mp_policy

            ignore_modules_in_mixed_precision = tuple()
            if hasattr(model, "get_ignore_modules_in_mixed_precision"):
                ignore_modules_in_mixed_precision = model.get_ignore_modules_in_mixed_precision()

            def apply_fsdp_to_decoder_blocks(module: "nn.Module") -> None:
                if module.__class__.__name__ in basic_modules or module.__class__ in ignore_modules_in_mixed_precision:
                    logger.debug(f"Apply FSDP2 to {module.__class__.__name__}.")
                    if module.__class__ in ignore_modules_in_mixed_precision:
                        fully_shard(module, **{k: v for k, v in fsdp_kwargs.items() if k != "mp_policy"})
                    else:
                        fully_shard(module, **fsdp_kwargs)

            if basic_modules:
                model.apply(apply_fsdp_to_decoder_blocks)

            fully_shard(model, **fsdp_kwargs)

            if kwargs.get("init_device") == "meta":
                if weights_path is None:
                    # shard init empty model with fsdp2
                    model.to_empty(device="cuda")
                    model.init_weights()
                else:
                    from torch.distributed.tensor import distribute_tensor

                    load_model_weights(model, weights_path, "cuda", dtensor_factory=distribute_tensor)

        elif parallel_state.dp_mode == "fsdp1":
            wrap_policy = partial(
                lambda_auto_wrap_policy, lambda_fn=lambda module: module.__class__.__name__ in basic_modules
            )

            # set fsdp/hsdp sharding strategy
            if parallel_state.fsdp_mesh.ndim > 1 and parallel_state.fsdp_mesh.size() > 1:
                strategy = ShardingStrategy.HYBRID_SHARD
            else:
                strategy = ShardingStrategy.FULL_SHARD

            fsdp_kwargs = {
                "auto_wrap_policy": wrap_policy,
                "ignored_states": fsdp_no_shard_states,
                "device_id": torch.cuda.current_device(),
                "sharding_strategy": strategy if enable_full_shard else ShardingStrategy.NO_SHARD,
                "use_orig_params": True,
            }

            fsdp_kwargs["device_mesh"] = parallel_state.fsdp_mesh

            fsdp_kwargs.update(kwargs.pop("fsdp_kwargs", {}))

            if enable_mixed_precision:
                logger.info_rank0("Enable mixed precision training.")
                mixed_precision = MixedPrecision(
                    param_dtype=torch.bfloat16,
                    reduce_dtype=torch.float32,
                    buffer_dtype=torch.float32,
                )
                if hasattr(model, "get_ignore_modules_in_mixed_precision"):
                    mixed_precision._module_classes_to_ignore += model.get_ignore_modules_in_mixed_precision()

                fsdp_kwargs["mixed_precision"] = mixed_precision

            if kwargs.get("init_device") == "cpu":
                logger.info_rank0("Enable rank0-only initialization.")
                fsdp_kwargs["sync_module_states"] = True
                if parallel_state.global_rank != 0:
                    fsdp_kwargs["param_init_fn"] = init_fsdp_fn(model, device="cuda")
            elif kwargs.get("init_device") == "meta":
                # assert weights_path is not None, "`weights_path` must be provided when `init_device=meta` for fsdp1."

                logger.info_rank0("Enable meta initialization.")
                if weights_path is None:
                    logger.info_rank0("weights_path is None during meta initialization.")

                ignore_param_names = (
                    [".".join([fqn, k]) for fqn in fsdp_no_shard_states_fqn for k in ep_param_suffix]
                    if fsdp_no_shard_states_fqn is not None
                    else None
                )
                shard_states = (
                    parallel_load_safetensors(weights_path, ignore_param_name=ignore_param_names)
                    if weights_path
                    else kwargs.get("state_dict", {})
                )
                fsdp_kwargs["param_init_fn"] = parallel_init_fsdp_fn(
                    model, shard_states, ignore_param_name=ignore_param_names
                )

            if kwargs.pop("enable_fsdp_offload", False):
                logger.info_rank0("Enable offloading for parameters & gradients & optimizer states.")
                fsdp_kwargs["cpu_offload"] = CPUOffload(offload_params=True)

            if kwargs.pop("enable_forward_prefetch", False):
                fsdp_kwargs["forward_prefetch"] = True
            else:
                fsdp_kwargs["forward_prefetch"] = False
                fsdp_kwargs["backward_prefetch"] = None

            # FULLY_SHARD first
            model = FullyShardedDataParallel(model, **fsdp_kwargs)

            if fsdp_no_shard_states is not None:
                # apply NO_SHARD the ignored_states, but wrap into DDP
                if parallel_state.ep_fsdp_mesh["ep_fsdp"].size() == 1:
                    moe_sharding_strategy = ShardingStrategy.NO_SHARD
                    ep_fsdp_device_mesh = parallel_state.fsdp_mesh
                else:
                    moe_sharding_strategy = ShardingStrategy.FULL_SHARD
                    ep_fsdp_device_mesh = parallel_state.ep_fsdp_mesh["ep_fsdp"]

                logger.info_rank0(f"Apply {moe_sharding_strategy} states on '{fsdp_no_shard_states_fqn}'.")
                fsdp_kwargs.pop("ignored_states", None)
                fsdp_kwargs.pop("auto_wrap_policy", None)
                fsdp_kwargs["sharding_strategy"] = moe_sharding_strategy
                fsdp_kwargs["device_mesh"] = ep_fsdp_device_mesh
                logger.info_rank0(f"{ep_fsdp_device_mesh=}")
                for fqn in fsdp_no_shard_states_fqn:
                    no_shard_module = get_module_from_path(model, fqn)
                    if kwargs.get("init_device") == "meta":
                        specific_param_name = [".".join([fqn, k]) for k in ep_param_suffix]
                        shard_states = (
                            parallel_load_safetensors(weights_path, specific_param_name=specific_param_name)
                            if weights_path
                            else {}
                        )
                        if weights_path:
                            for suffix in ep_param_suffix:
                                shard_states[suffix] = shard_states.pop(".".join([fqn, suffix]))
                        fsdp_kwargs["param_init_fn"] = parallel_init_fsdp_fn(
                            no_shard_module, shard_states, specific_param_name=ep_param_suffix
                        )
                    fsdp_module = FullyShardedDataParallel(no_shard_module, **fsdp_kwargs)
                    fsdp_state = _get_module_fsdp_state_if_fully_sharded_module(fsdp_module)
                    fsdp_state._gradient_postdivide_factor *= parallel_state.ep_size
                    set_module_from_path(model, fqn, fsdp_module)

            _lazy_init(model, model)

            # Apply fsdp extension to FSDP model
            save_hook_mesh = parallel_state.ep_fsdp_device_mesh if parallel_state.ep_enabled else None
            logger.info_rank0("Register Checkpoints Extension hook to the model")
            register_checkpoint_extension(
                fsdp_model=model,
                save_hook_mesh=save_hook_mesh,
                fqn2spec_info=fqn2spec_info,
            )

            if parallel_state.ep_enabled:
                model.clip_grad_norm_ = types.MethodType(clip_grad_norm_, model)

            verbose_fsdp_grouping(model)
        else:
            ddp_kwargs = {"device_ids": [parallel_state.local_rank]}
            if enable_mixed_precision:
                logger.info_rank0("Enable mixed precision training.")
                mixed_precision = MixedPrecision(
                    param_dtype=torch.bfloat16,
                    reduce_dtype=torch.float32,
                    buffer_dtype=torch.bfloat16,
                )
                ddp_kwargs["mixed_precision"] = mixed_precision

            model = DDP(model, **ddp_kwargs)

    return model



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Dict


class DynBszBuffer:
    """
    A buffer to store samples for dynamic batch size.
    """

    def __init__(self):
        self._buffer = []
        self._buffer_sample_lens = []
        self.del_idxs = []
        self.cur_idx = 0
        self.all_token_cnt = 0

    def append(self, item: Dict[str, Any]):
        """
        Append a sample to the buffer.
        Args:
            item: a sample to append to the buffer.
                The sample should be a dict with the following keys:
                    - input_ids: torch.Tensor of shape (seq_len, )
                    - attention_mask: torch.Tensor of shape (seq_len, )
        """
        self._buffer.append(item)
        self._buffer_sample_lens.append(item["attention_mask"].sum())
        self.all_token_cnt += self._buffer_sample_lens[-1]

    def get_samples(self, n_token_per_iter: int, force: bool = True):
        """
        get samples from the buffer.
        Args:
            n_token_per_iter: the number of tokens to get.
            force: if True, the first sample will be returned even if it is not full.
        Returns:
            samples: a list of samples.
        """
        cum_seq_len = 0
        samples = []
        while self.cur_idx < len(self._buffer) and cum_seq_len < n_token_per_iter:
            seq_len = self._buffer_sample_lens[self.cur_idx]
            if self.cur_idx not in self.del_idxs and (
                (force is True and cum_seq_len == 0) or (seq_len <= n_token_per_iter - cum_seq_len)
            ):
                cum_seq_len += seq_len
                samples.append(self._buffer[self.cur_idx])
                self.del_idxs.append(self.cur_idx)
            self.cur_idx += 1
        assert len(samples) > 0
        return samples

    def __len__(self):
        return len(self._buffer)

    def flush(self):
        """ "
        Flush the buffer.
        """
        self.cur_idx = 0
        self.all_token_cnt -= sum([self._buffer_sample_lens[idx] for idx in self.del_idxs])
        buffer_len = len(self._buffer)
        self._buffer = [self._buffer[idx] for idx in range(buffer_len) if idx not in self.del_idxs]
        self._buffer_sample_lens = [
            self._buffer_sample_lens[idx] for idx in range(buffer_len) if idx not in self.del_idxs
        ]
        self.del_idxs = []

    def merge(self, buffer_to_merge: "DynBszBuffer"):
        """ "
        Merge the buffer with another buffer.
        Args:
            buffer_to_merge: the buffer to merge.
        """
        self.flush()
        buffer_to_merge.flush()
        for item in buffer_to_merge._buffer:
            self.append(item)


class BaseBatchingStrategy:
    """
    Base class for batching strategy.s
    """

    def is_full_filled(self) -> bool:
        raise NotImplementedError("should implement `is_full_filled`")

    def put_item(self, item: Dict[str, Any]):
        raise NotImplementedError("should implement `put_item`")

    def get_micro_batch(self, step: int) -> Any:
        raise NotImplementedError("should implement `get_micro_batch` ")

    def empty(self) -> bool:
        raise NotImplementedError("should implement `empty`")


class IdentityPacker:
    def __init__(self, token_micro_bsz, bsz_warmup_steps, bsz_warmup_init_mbtoken):
        self.token_micro_bsz = token_micro_bsz
        self.bsz_warmup_steps = bsz_warmup_steps
        self.bsz_warmup_init_mbtoken = bsz_warmup_init_mbtoken

    def __call__(self, samples):
        return samples

    def get_token_num_to_request(self, cur_step, warmup):
        return (
            (self.token_micro_bsz - self.bsz_warmup_init_mbtoken) * cur_step // self.bsz_warmup_steps
            + self.bsz_warmup_init_mbtoken
            if warmup
            else self.token_micro_bsz
        )


class TextBatchingStrategy(BaseBatchingStrategy):
    """ "
    Batching strategy for text data.
    Args:
        token_micro_bsz: the number of tokens to get for each request.
        buffer_size: the size of the buffer.
        bsz_warmup_steps: the number of steps to warm up the batch size.
        bsz_warmup_init_mbtoken: the initial number of tokens to get for each request.
    """

    def __init__(
        self,
        token_micro_bsz,
        buffer_size: int = 500,
        bsz_warmup_steps: int = -1,
        bsz_warmup_init_mbtoken: int = 200,
    ) -> None:
        super().__init__()
        self._step = 0
        self.token_micro_bsz = token_micro_bsz
        self.bsz_warmup_steps = bsz_warmup_steps
        self.buffer_size = buffer_size  # minimum samples in buffer
        self.buffer = DynBszBuffer()
        self.bsz_warmup_init_mbtoken = bsz_warmup_init_mbtoken
        assert self.bsz_warmup_init_mbtoken >= 0

        self.packer = IdentityPacker(
            token_micro_bsz=token_micro_bsz,
            bsz_warmup_steps=bsz_warmup_steps,
            bsz_warmup_init_mbtoken=bsz_warmup_init_mbtoken,
        )

    def is_full_filled(self) -> bool:
        return len(self.buffer) >= self.buffer_size and self.buffer.all_token_cnt >= self.token_micro_bsz

    def put_item(self, item: Dict[str, Any]):
        if len(item["input_ids"]) == 1:
            print("WARNING: EMPTY STRING.")
            return
        self.buffer.append(item)

    def get_token_num_to_request(self):
        if self.packer is not None:
            warmup = self._step <= self.bsz_warmup_steps and self.bsz_warmup_steps > 0
            return self.packer.get_token_num_to_request(self._step, warmup=warmup)
        else:
            return self.get_cur_token_micro_bsz()

    def get_cur_token_micro_bsz(self):
        warmup = self._step <= self.bsz_warmup_steps and self.bsz_warmup_steps > 0
        if warmup:
            return (
                self.token_micro_bsz - self.bsz_warmup_init_mbtoken
            ) * self._step // self.bsz_warmup_steps + self.bsz_warmup_init_mbtoken
        else:
            return self.token_micro_bsz

    def get_micro_batch(self, step) -> Any:
        """
        Get a micro batch from the buffer according to the current step.
        Args:
            step: the current step.
        Returns:
            data: a list of samples.
        """

        self._step = step
        n_token_per_iter = self.get_token_num_to_request()
        cur_token_micro_bsz = self.get_cur_token_micro_bsz()
        assert cur_token_micro_bsz % n_token_per_iter == 0, (
            "The token num to get for each request should be divisible by token micro bsz."
        )
        n_iter = int(cur_token_micro_bsz // n_token_per_iter)
        data = []
        for i in range(n_iter):
            samples = self.buffer.get_samples(n_token_per_iter)
            if self.packer:
                samples = self.packer(samples)  # maybe packed into one sample, but wrapped in list.
            data.extend(samples)
        self.buffer.flush()  # remove the selected samples.
        return data

    def empty(self) -> bool:
        return len(self.buffer) == 0



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os
from collections import OrderedDict
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Literal, Optional, Sequence, Tuple, Union

import torch
from diffusers.utils import SAFE_WEIGHTS_INDEX_NAME as DIFFUSERS_SAFE_WEIGHTS_INDEX_NAME
from diffusers.utils import SAFETENSORS_WEIGHTS_NAME as DIFFUSERS_SAFETENSORS_WEIGHTS_NAME
from torch import distributed as dist
from torch import nn
from tqdm import tqdm
from transformers.utils import SAFE_WEIGHTS_INDEX_NAME, SAFE_WEIGHTS_NAME, WEIGHTS_INDEX_NAME, WEIGHTS_NAME
from transformers.utils.hub import cached_file, get_checkpoint_shard_files
from transformers.utils.import_utils import is_safetensors_available

from ..utils import logging
from ..utils.helper import empty_cache, get_dtype_size


if is_safetensors_available():
    from safetensors import safe_open
    from safetensors.torch import save_file


if TYPE_CHECKING:
    from transformers import GenerationConfig, PretrainedConfig, PreTrainedModel, PreTrainedTokenizer, ProcessorMixin

    ModelAssets = Union[GenerationConfig, PretrainedConfig, PreTrainedTokenizer, ProcessorMixin]


logger = logging.get_logger(__name__)


@contextmanager
def init_empty_weights():
    """
    A context manager under which models are initialized with all parameters on the meta device.

    Borrowed from: https://github.com/huggingface/accelerate/blob/v1.0.0rc1/src/accelerate/big_modeling.py#L57
    """
    old_register_parameter = nn.Module.register_parameter

    def register_empty_parameter(module: "nn.Module", name: str, param: "nn.Parameter"):
        old_register_parameter(module, name, param)
        if param is not None:
            param_cls = type(module._parameters[name])
            kwargs = module._parameters[name].__dict__
            kwargs["requires_grad"] = param.requires_grad
            module._parameters[name] = param_cls(module._parameters[name].to("meta"), **kwargs)

    try:
        nn.Module.register_parameter = register_empty_parameter
        yield
    finally:
        nn.Module.register_parameter = old_register_parameter


@dataclass
class StateDictIterator:
    filepath: str

    def __iter__(self) -> Generator[Tuple[str, "torch.Tensor"], None, None]:
        if self.filepath.endswith(".safetensors"):
            with safe_open(self.filepath, framework="pt", device="cpu") as f:
                for key in f.keys():
                    yield key, f.get_tensor(key)

        else:
            state_dict = torch.load(self.filepath, map_location="cpu", weights_only=True, mmap=True)
            for key in state_dict.keys():
                yield key, state_dict[key]


def _load_state_dict(weights_path: str, **kwargs) -> List["StateDictIterator"]:
    """
    Loads (sharded) state dict in transformers' format.
    """
    cache_kwargs = {"_raise_exceptions_for_missing_entries": False, **kwargs}
    resolved_weight_file = cached_file(weights_path, SAFE_WEIGHTS_NAME, **cache_kwargs)
    if resolved_weight_file:
        return [StateDictIterator(resolved_weight_file)]

    resolved_weight_file = cached_file(weights_path, SAFE_WEIGHTS_INDEX_NAME, **cache_kwargs)
    if resolved_weight_file:
        shard_files, _ = get_checkpoint_shard_files(weights_path, resolved_weight_file, **kwargs)
        return [StateDictIterator(shard_file) for shard_file in shard_files]

    resolved_weight_file = cached_file(weights_path, DIFFUSERS_SAFETENSORS_WEIGHTS_NAME, **cache_kwargs)
    if resolved_weight_file:
        return [StateDictIterator(resolved_weight_file)]

    resolved_weight_file = cached_file(weights_path, DIFFUSERS_SAFE_WEIGHTS_INDEX_NAME, **cache_kwargs)
    if resolved_weight_file:
        shard_files, _ = get_checkpoint_shard_files(weights_path, resolved_weight_file, **kwargs)
        return [StateDictIterator(shard_file) for shard_file in shard_files]

    resolved_weight_file = cached_file(weights_path, WEIGHTS_NAME, **cache_kwargs)
    if resolved_weight_file:
        return [StateDictIterator(resolved_weight_file)]

    resolved_weight_file = cached_file(weights_path, WEIGHTS_INDEX_NAME, **cache_kwargs)
    if resolved_weight_file:
        shard_files, _ = get_checkpoint_shard_files(weights_path, resolved_weight_file, **kwargs)
        return [StateDictIterator(shard_file) for shard_file in shard_files]

    raise ValueError(f"Cannot find checkpoint files in {weights_path}.")


def _find_submodule(module: "nn.Module", name: str) -> Tuple["nn.Module", str]:
    """
    Finds the leaf module according to the name.
    """
    pieces = name.split(".")
    for piece in pieces[:-1]:
        if not hasattr(module, piece):
            raise ValueError(f"Cannot find {piece} in {module}.")

        module = getattr(module, piece)

    return module, pieces[-1]


def _dispatch_parameter(
    module: "nn.Module",
    name: str,
    tensor: "torch.Tensor",
    dtensor_factory: Optional[Callable[["torch.Tensor", Any, Any], "torch.Tensor"]] = None,
) -> None:
    """
    Assigns parameter to an empty model.

    NOTE: FSDP module must use in-place operators.
    """
    module, name = _find_submodule(module, name)
    orig_tensor = module._parameters[name].data
    tensor = tensor.to(orig_tensor)
    if hasattr(orig_tensor, "device_mesh"):  # dtensor
        if orig_tensor.device.type == "cpu":
            raise ValueError("Cannot load dtensor on CPU.")

        device_mesh = getattr(orig_tensor, "device_mesh")
        placements = getattr(orig_tensor, "placements")
        module._parameters[name].data.copy_(dtensor_factory(tensor, device_mesh, placements))
    else:  # not dtensor
        module._parameters[name].data.copy_(tensor)


def _dispatch_buffer(
    module: "nn.Module",
    name: str,
    buffer: "torch.Tensor",
) -> None:
    """
    Assigns buffer to an empty model.
    """
    module, name = _find_submodule(module, name)
    orig_tensor = module._buffers[name].data
    module._buffers[name] = buffer.to(orig_tensor)


def _init_parameter(
    module: "nn.Module",
    name: str,
) -> None:
    """
    Initializes parameter in model.
    """
    pieces = name.split(".")
    init_func = None
    for piece in pieces[:-1]:
        if not hasattr(module, piece):
            raise ValueError(f"Cannot find {piece} in {module}.")

        if hasattr(module, "_init_weights"):
            init_func = getattr(module, "_init_weights")

        module = getattr(module, piece)

    if init_func is None:
        raise ValueError(f"Cannot retrieve `_init_weights` function in the parents of {module}.")

    module.apply(init_func)


@torch.no_grad()
def load_model_weights(
    model: Union["nn.Module", "PreTrainedModel"],
    weights_path: str,
    init_device: Literal["cpu", "cuda"] = "cuda",
    dtensor_factory: Optional[Callable[["torch.Tensor", Any, Any], "torch.Tensor"]] = None,
) -> None:
    """
    Loads pre-trained model states in transformers' format.
    """
    buffer_dict = {name: buffer.clone() for name, buffer in model.named_buffers()}
    parameter_names = {name for name, _ in model.named_parameters()}
    model.to_empty(device=init_device)
    state_dict_iterators = _load_state_dict(weights_path)
    for state_dict_iterator in tqdm(
        state_dict_iterators, desc="Loading checkpoint shards", disable=int(os.getenv("LOCAL_RANK", "-1")) > 0
    ):
        for name, tensor in state_dict_iterator:
            if name in buffer_dict.keys():  # persistent buffers
                buffer_dict[name] = tensor.clone()
            elif name in parameter_names:
                parameter_names.remove(name)
                _dispatch_parameter(model, name, tensor, dtensor_factory)
            else:
                logger.info_rank0(f"Unexpected key in state dict: {name}.")

        del state_dict_iterator
        empty_cache()

    for name, buffer in buffer_dict.items():
        _dispatch_buffer(model, name, buffer)

    if len(parameter_names) > 0:
        logger.info_rank0(f"Find missing key(s) in state dict: {parameter_names}, initialize them.")
        for name in parameter_names:
            _init_parameter(model, name)

    # we should tie embeddings after loading weights because to_empty() leads to untied weights,
    # except for fsdp1 (custom init) and fsdp2 (swap tensor) contexts.
    if getattr(model.config, "tie_word_embeddings", True):
        try:
            input_embeddings = model.get_input_embeddings()
            output_embeddings = model.get_output_embeddings()
            output_embeddings._parameters["weight"] = input_embeddings._parameters["weight"]
        except Exception as e:
            logger.info_rank0(f"Failed to tie embeddings: {e}")


def _get_shard_info(
    state_dict: Dict[str, "torch.Tensor"],
    save_dtype: Optional[Union[str, "torch.dtype"]],
    shard_size: int,
    safe_serialization: bool,
) -> Tuple[bool, int, Dict[str, str]]:
    """
    Gets the shard information, should be executed at rank 0.
    """
    current_size, total_size = 0, 0
    current_shard, shard_list = [], []
    for name, tensor in state_dict.items():
        if isinstance(save_dtype, str):
            dtype = getattr(torch, save_dtype)
        elif isinstance(save_dtype, torch.dtype):
            dtype = save_dtype
        else:
            dtype = tensor.dtype
        tensor_size = tensor.numel() * get_dtype_size(dtype)  # dtensor's numel == tensor's numel
        if current_size != 0 and current_size + tensor_size > shard_size:
            total_size += current_size
            shard_list.append(current_shard)
            current_size = 0
            current_shard = []

        current_size += tensor_size
        current_shard.append(name)

    if current_size != 0:
        total_size += current_size
        shard_list.append(current_shard)

    weights_name = SAFE_WEIGHTS_NAME if safe_serialization else WEIGHTS_NAME
    num_shards = len(shard_list)
    weight_map = OrderedDict()
    is_sharded = None
    if num_shards == 1:
        is_sharded = False
        for name in shard_list[0]:
            weight_map[name] = weights_name
    else:
        is_sharded = True
        for shard_idx, shard in enumerate(shard_list):
            prefix, extension = weights_name.rsplit(".", maxsplit=1)
            file_name = f"{prefix}-{shard_idx + 1:05d}-of-{num_shards:05d}.{extension}"
            for name in shard:
                weight_map[name] = file_name

    return is_sharded, total_size, weight_map


def _save_state_dict(
    state_dict: Dict[str, "torch.Tensor"],
    path_to_save: "os.PathLike",
    safe_serialization: bool,
) -> None:
    """
    Save function.
    """
    if safe_serialization:
        save_file(state_dict, path_to_save, metadata={"format": "pt"})
    else:
        torch.save(state_dict, path_to_save)


@torch.no_grad()
def save_model_weights(
    output_dir: Union[str, "os.PathLike"],
    state_dict: Dict[str, "torch.Tensor"],
    global_rank: Optional[int] = None,
    save_dtype: Optional[Union[str, "torch.dtype"]] = "bfloat16",
    shard_size: int = 5_000_000_000,
    safe_serialization: bool = True,
    model_assets: Optional[Sequence["ModelAssets"]] = None,
) -> None:
    """
    Saves full model weights. The model parameters should be either tensor or dtensor.

    If global_rank is given, it will assume it is executed on all ranks.
    """

    os.makedirs(output_dir, exist_ok=True)
    is_sharded, total_size, weight_map = _get_shard_info(state_dict, save_dtype, shard_size, safe_serialization)
    full_state_dict = OrderedDict()
    prev_file_name = None
    for name, tensor in state_dict.items():
        if hasattr(tensor.data, "full_tensor"):  # dtensor
            tensor = tensor.data.full_tensor()
        else:
            tensor = tensor.data

        if save_dtype:
            tensor = tensor.to(dtype=getattr(torch, save_dtype) if isinstance(save_dtype, str) else save_dtype)

        if prev_file_name is not None and weight_map[name] != prev_file_name:
            if global_rank is None or global_rank == 0:
                _save_state_dict(full_state_dict, os.path.join(output_dir, prev_file_name), safe_serialization)
                full_state_dict = OrderedDict()

            empty_cache()
            if global_rank is not None and dist.is_initialized():  # avoid process hanging
                torch.cuda.synchronize()
                dist.barrier()

        if global_rank is None or global_rank == 0:
            full_state_dict[name] = tensor.detach().cpu()

        prev_file_name = weight_map[name]
        del tensor

    if global_rank is None or global_rank == 0:
        if len(full_state_dict):
            _save_state_dict(full_state_dict, os.path.join(output_dir, prev_file_name), safe_serialization)

        if is_sharded:
            index = {
                "metadata": {"total_size": total_size},
                "weight_map": weight_map,
            }

            index_file = SAFE_WEIGHTS_INDEX_NAME if safe_serialization else WEIGHTS_INDEX_NAME
            with open(os.path.join(output_dir, index_file), "w", encoding="utf-8") as f:
                content = json.dumps(index, indent=2, sort_keys=True) + "\n"
                f.write(content)

            logger.info(f"Model weight splits saved in {output_dir}.")
        else:
            logger.info(f"Model weights saved at {os.path.join(output_dir, prev_file_name)}.")

        if model_assets is not None:
            for model_asset in model_assets:
                if hasattr(model_asset, "save_pretrained"):
                    model_asset.save_pretrained(output_dir)
                else:
                    logger.warning(f"Model asset {model_asset} should implement `save_pretrained`.")


def save_model_assets(output_dir: Union[str, "os.PathLike"], model_assets: Sequence["ModelAssets"]):
    for model_asset in model_assets:
        if hasattr(model_asset, "save_pretrained"):
            model_asset.save_pretrained(output_dir)
        else:
            logger.warning(f"Model asset {model_asset} should implement `save_pretrained`.")



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from transformers import ViTConfig, ViTModel


def build_feature_projector(
    feature_size: int,
    hidden_size: int,
    mlp_depth: int = 2,
) -> "torch.nn.Module":
    """
    Builds feature projector containing several linear layers.
    """
    layers = [torch.nn.Linear(feature_size, hidden_size)]
    for _ in range(mlp_depth - 1):
        layers.append(torch.nn.GELU())
        layers.append(torch.nn.Linear(hidden_size, hidden_size))

    return torch.nn.Sequential(*layers)


def build_vit_decoder(
    input_dim: int,
    num_layers: int,
    num_heads: int,
    hidden_dim: int,
    output_dim: int,
):
    class ViT(torch.nn.Module):
        def __init__(self):
            super().__init__()
            vit_config = ViTConfig(
                hidden_size=hidden_dim,
                num_hidden_layers=num_layers,
                num_attention_heads=num_heads,
                intermediate_size=hidden_dim * 4,
                image_size=32,
                patch_size=1,
                num_channels=input_dim,  # Input feature channel size
            )
            self.projector = torch.nn.Linear(hidden_dim, output_dim)
            self.model = ViTModel(vit_config)

        def forward(self, x):
            x = self.model(x).last_hidden_state
            x = x[:, 1:, :]  # skip cls_token
            return self.projector(x)

    return ViT()



from typing import Optional, Tuple

import torch
from transformers.modeling_flash_attention_utils import _flash_attention_forward

from ..distributed.parallel_state import get_parallel_state
from ..distributed.sequence_parallel import (
    gather_heads_scatter_seq,
    gather_seq_scatter_heads,
)
from ..utils import logging


logger = logging.get_logger(__name__)


def repeat_kv(hidden_states: torch.Tensor, n_rep: int) -> torch.Tensor:
    """
    This is the equivalent of torch.repeat_interleave(x, dim=1, repeats=n_rep). The hidden states go from (batch,
    num_key_value_heads, seqlen, head_dim) to (batch, num_attention_heads, seqlen, head_dim)
    """
    batch, num_key_value_heads, slen, head_dim = hidden_states.shape
    if n_rep == 1:
        return hidden_states
    hidden_states = hidden_states[:, :, None, :, :].expand(batch, num_key_value_heads, n_rep, slen, head_dim)
    return hidden_states.reshape(batch, num_key_value_heads * n_rep, slen, head_dim)


def flash_attention_forward(
    module: torch.nn.Module,
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attention_mask: Optional[torch.Tensor],
    dropout: float = 0.0,
    scaling: Optional[float] = None,
    sliding_window: Optional[int] = None,
    softcap: Optional[float] = None,
    **kwargs,
) -> Tuple[torch.Tensor, None]:
    if kwargs.get("output_attentions", False) or kwargs.get("head_mask", None) is not None:
        logger.warning_once(
            "`flash_attention_2` does not support `output_attentions=True` or `head_mask`."
            " Please set your attention to `eager` if you want any of these features."
        )

    # FA2 uses non-transposed inputs
    query = query.transpose(1, 2)
    key = key.transpose(1, 2)
    value = value.transpose(1, 2)

    # FA2 always relies on the value set in the module, so remove it if present in kwargs to avoid passing it twice
    kwargs.pop("is_causal", None)

    # This is for Qwen2VL's mrope
    position_ids = kwargs.pop("position_ids", None)
    if position_ids is not None and position_ids.dim() == 3:
        position_ids = position_ids[0]

    # Ulysses patch
    ulysses_enabled = get_parallel_state().ulysses_enabled
    if ulysses_enabled:
        ulysses_group = get_parallel_state().ulysses_group
        # Sanity Check & Repeat Key & Value
        ulysses_size = get_parallel_state().ulysses_size
        q_head_num = query.shape[2]
        kv_head_num = key.shape[2]
        unpadded_seq_len = None

        assert q_head_num % ulysses_size == 0, (
            f"num_query_heads ({q_head_num}) must be divisible by ulysses_size ({ulysses_size})"
        )
        if ulysses_size > kv_head_num:
            assert ulysses_size % kv_head_num == 0, (
                f"ulysses_size ({ulysses_size}) must be divisible by num_key_value_heads ({kv_head_num})"
            )
            n_repeat = ulysses_size // kv_head_num
            key = repeat_kv(key, n_repeat)
            value = repeat_kv(value, n_repeat)

        if query.ndim == 4 and query.size(0) == 1:
            query, key, value = query.squeeze(0), key.squeeze(0), value.squeeze(0)
            query = gather_seq_scatter_heads(
                query, seq_dim=0, head_dim=1, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            key = gather_seq_scatter_heads(
                key, seq_dim=0, head_dim=1, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            value = gather_seq_scatter_heads(
                value, seq_dim=0, head_dim=1, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            query, key, value = query.unsqueeze(0), key.unsqueeze(0), value.unsqueeze(0)
        else:
            query = gather_seq_scatter_heads(
                query, seq_dim=1, head_dim=2, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            key = gather_seq_scatter_heads(
                key, seq_dim=1, head_dim=2, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            value = gather_seq_scatter_heads(
                value, seq_dim=1, head_dim=2, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )

    # Only after all_to_all we got the full seq_len
    seq_len = query.shape[1]

    attn_output: torch.Tensor = _flash_attention_forward(
        query,
        key,
        value,
        attention_mask,
        query_length=seq_len,
        is_causal=module.is_causal,
        dropout=dropout,
        position_ids=position_ids,
        softmax_scale=scaling,
        sliding_window=sliding_window,
        softcap=softcap,
        use_top_left_mask=False,
        **kwargs,
    )

    # Ulysses patch
    if ulysses_enabled:
        ulysses_group = get_parallel_state().ulysses_group
        if attn_output.ndim == 4 and attn_output.size(0) == 1:
            attn_output = attn_output.squeeze(0)
            attn_output = gather_heads_scatter_seq(attn_output, seq_dim=0, head_dim=1, group=ulysses_group)
            attn_output = attn_output.unsqueeze(0)
        else:
            attn_output = gather_heads_scatter_seq(attn_output, seq_dim=1, head_dim=2, group=ulysses_group)

    return attn_output, None



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
from functools import partial
from typing import Any, Dict, List, Tuple, Union

import torch
from torch.distributed._tensor import DeviceMesh, DTensor, Replicate, Shard
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp._fsdp_extensions import FSDPExtensions
from torch.distributed.fsdp._optim_utils import (
    FSDPParamInfo,
    _OptimStateKey,
    _unflatten_optim_state,
    sorted_items,
)

from ...utils import logging
from ...utils.import_utils import is_torch_version_greater_than
from ..parallel_plan import SpecInfo


logger = logging.get_logger(__name__)

OPTIM_STATE_NO_SHARD_KEY = ["step"]
orig_optim_state_dict = FSDP.optim_state_dict
orig_optim_state_dict_to_load = FSDP.optim_state_dict_to_load


def _shard_tensor(orgin_tensor: torch.Tensor, device_mesh: DeviceMesh, shard: Shard = Shard(0)):
    """
    Shard Tensor to DTensor.

    args:
        orgin_tensor (torch.Tensor): The orgin tensor.
        device_mesh (DeviceMesh): The ep device mesh.
        shard (Shard): The shard info, default Shard(0).

    """
    assert device_mesh.ndim == 2, f"global_mesh.ndim must be 2, got {device_mesh.ndim}"
    ep_mesh = device_mesh["ep"]

    if orgin_tensor.__class__.__name__ == "DTensor":
        dtensor = DTensor.from_local(orgin_tensor._local_tensor, device_mesh=device_mesh, placements=[shard, shard])
    elif orgin_tensor.__class__.__name__ == "Tensor":
        dtensor = DTensor.from_local(orgin_tensor, device_mesh=ep_mesh, placements=[shard])

    return dtensor


def _shard_dtensor(orgin_dtensor: DTensor, device_mesh: DeviceMesh, shard: Shard = Shard(0)):
    """
    Convert DTensor to local Tensor

    args:
        orgin_dtensor (torch.Tensor): The orgin tensor.
        device_mesh (DeviceMesh): The ep device mesh.
        shard (Shard): The shard info, default Shard(0).

    """
    assert isinstance(orgin_dtensor, DTensor), (
        f"Only support DTensor, got {type(orgin_dtensor)}, for torch.Tensor, use _shard_dtensor instead."
    )

    local_tensor = orgin_dtensor.to_local()

    return local_tensor


def check_any_unflat_param_names_match(unflat_param_name: str, fqn2spec_info: Dict[str, SpecInfo], prefix: str = None):
    assert isinstance(unflat_param_name, str), f"unflat_param_name must be a str, got {type(unflat_param_name)}"

    if prefix:
        assert unflat_param_name.startswith(prefix), (
            f"unflat_param_name {unflat_param_name} must start with prefix {prefix}"
        )
        unflat_param_name = unflat_param_name[len(prefix) :].lstrip(".")

    if unflat_param_name not in fqn2spec_info:
        logger.warning_rank0(f"unflat_param_name {unflat_param_name} not in fqn2spec_info.")
        return False

    if isinstance(fqn2spec_info[unflat_param_name].placement, Shard):
        return True

    return False


def check_all_unflat_param_names_match(unflat_param_names: Tuple[str], fqn2spec_info: Dict[str, SpecInfo]):
    """
    Check
    """
    assert isinstance(unflat_param_names, (list, tuple)), (
        f"unflat_param_names must be a list or tuple, got {type(unflat_param_names)}"
    )

    unflat_len = len(unflat_param_names)
    cnt = 0
    for names in unflat_param_names:
        assert names in fqn2spec_info, (
            f"unflat_param_names {unflat_param_names} must be in fqn2spec_info {fqn2spec_info}"
        )
        if isinstance(fqn2spec_info[names].placement, Shard):
            cnt += 1
    assert cnt == 0 or cnt == unflat_len, f"unflat_param_names {unflat_param_names} must be all shard or all not shard"

    return cnt == unflat_len


class CheckpointExtensions(FSDPExtensions):
    def __init__(
        self,
        ep_fsdp_device_mesh: DeviceMesh,
        fqn2spec_info: Dict[str, SpecInfo],
    ):
        super().__init__()
        self.ep_fsdp_device_mesh = ep_fsdp_device_mesh
        self.ep_mesh = ep_fsdp_device_mesh["ep"] if ep_fsdp_device_mesh is not None else None
        self.fqn2spec_info = fqn2spec_info

    def chunk_dtensor(self, tensor: torch.Tensor, rank: int, device_mesh: DeviceMesh) -> torch.Tensor:
        """Shards a tensor/DTensor to DTensor and returns the local DTensor."""
        # We need to explicitly call .detach() to return a new tensor detached from the current graph.
        tensor = tensor.clone().detach()
        fsdp_size = device_mesh.size(-1)
        dimlens = tuple(tensor.size())
        # by default we use the max-len dimension for sharding
        selected_dim = dimlens.index(max(dimlens))
        for dim, dimlen in enumerate(dimlens):
            if dimlen % fsdp_size == 0:
                selected_dim = dim
                break
        # HSDP placements: [Replicate(), ..., Shard(selected_dim)]
        replicate_placements = [Replicate() for _ in range(device_mesh.ndim)]
        shard_placements = [Replicate() for _ in range(device_mesh.ndim)]
        shard_placements[-1] = Shard(selected_dim)  # type: ignore[call-overload]
        dtensor = DTensor.from_local(tensor, device_mesh, replicate_placements, run_check=False).redistribute(
            placements=shard_placements,
        )

        return dtensor

    def chunk_tensor(self, tensor, rank, world_size, num_devices_per_node, pg, device=None):
        # use default
        raise NotImplementedError("Please init FSDP with device mesh")
        from torch.distributed.fsdp._fsdp_extensions import _ext_chunk_tensor

        return _ext_chunk_tensor(tensor, rank, world_size, num_devices_per_node, pg)

    def pre_flatten_transform(self, tensor):
        # use default
        from torch.distributed.fsdp._fsdp_extensions import _ext_pre_flatten_transform

        return _ext_pre_flatten_transform(tensor)

    def pre_load_state_dict_transform(self, tensor):
        # use default
        from torch.distributed.fsdp._fsdp_extensions import _ext_pre_load_state_dict_transform

        return _ext_pre_load_state_dict_transform(tensor)

    def post_unflatten_transform(self, tensor, param_extension):
        # use default
        from torch.distributed.fsdp._fsdp_extensions import _ext_post_unflatten_transform

        return _ext_post_unflatten_transform(tensor, param_extension)

    def all_gather_dtensor(self, tensor: DTensor, parent_mesh):
        # this is required during loading checkpoint (model.load_state_dict)
        # use default
        from torch.distributed.fsdp._fsdp_extensions import _ext_all_gather_dtensor

        if is_torch_version_greater_than("2.5.0"):
            return _ext_all_gather_dtensor(tensor, tensor.device_mesh)
        else:
            return _ext_all_gather_dtensor(tensor, None)

    @torch.no_grad()
    def state_dict_post_hook(
        self, module, state_dict, prefix, local_metadata, fqn2spec_info: Dict[str, SpecInfo] = None
    ):
        """
        Post state dict when calling `model.state_dict()` for EP cases.

        This will append EP placements to the FSDP DTensor state dicts
        """
        assert fqn2spec_info is not None, "if fqn2spec_info is None it should not be patch"

        if self.ep_mesh is None:
            return
        # [pp, ep_dp, ep, tp]
        global_device_mesh = self.ep_fsdp_device_mesh
        assert global_device_mesh.ndim == 2

        keys = list(state_dict.keys())
        for name in sorted(keys):
            if name in fqn2spec_info and isinstance(fqn2spec_info[name].placement, Shard):
                cur_spec_info = fqn2spec_info[name]
                tensor = state_dict[name]
                tensor = _shard_tensor(tensor, cur_spec_info.ep_fsdp_mesh, cur_spec_info.placement)
                state_dict[name] = tensor

    @torch.no_grad()
    def load_state_dict_pre_hook(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
        fqn2spec_info: Dict[str, SpecInfo] = None,
    ):
        """
        Pre load state dict when calling `model.load_state_dict()` for EP cases.

        This will shard Dtensor from ckpt to tensor state dicts
        """
        assert fqn2spec_info is not None, "if fqn2spec_info is None it should not be patch"

        if self.ep_mesh is None:
            return
        # [ep, fsdp-ep]
        global_device_mesh = self.ep_fsdp_device_mesh
        assert global_device_mesh.ndim == 2

        if self.ep_mesh.size() != global_device_mesh.size():
            return

        keys = list(state_dict.keys())
        for name in sorted(keys):
            tensor = state_dict[name]
            if check_any_unflat_param_names_match(name, fqn2spec_info, "_fsdp_wrapped_module"):
                fqn = name.split("_fsdp_wrapped_module.")[-1]
                cur_spec_info = fqn2spec_info[fqn]
                tensor = _shard_dtensor(tensor, cur_spec_info.ep_fsdp_mesh, cur_spec_info.placement)
                state_dict[name] = tensor

    def patch_convert_state_with_flat_params(self):
        """ """

        # Modify from torch.distributed.fsdp._optim_utils._convert_state_with_flat_params
        def _convert_state_with_flat_params_patch(
            all_optim_state_keys: List[_OptimStateKey],
            optim_state_key_to_param_key: Dict[_OptimStateKey, Union[int, str]],
            fqn_to_fsdp_param_info: Dict[str, FSDPParamInfo],
            optim_state_dict: Dict[Union[str, int], Any],
            to_save: bool,
            shard_state: bool,
            cpu_offload: bool = True,
            fqn2spec_info: Dict[str, SpecInfo] = None,
        ) -> Dict[str, Any]:
            fsdp_osd_state: Dict[str, Any] = {}
            # Iterate in rank 0's flat parameter ID order to ensure aligned all-gathers
            # across ranks
            for optim_state_key in all_optim_state_keys:
                param_key: Union[str, int, None] = optim_state_key_to_param_key.get(optim_state_key, None)

                assert param_key is not None, (
                    "If use_orig_params is False, we must be able to find the "
                    f"corresponding param id. {optim_state_key} {param_key}"
                )

                if optim_state_key.is_fsdp_managed:
                    # If there are multiple unflat_param_names (not use_orig_params),
                    # they share the same FSDPParamInfo. So the first unflat_param_name
                    # is sufficient to fetch the FSDPParamInfo.
                    fqn = optim_state_key.unflat_param_names[0]
                    fsdp_param_info = fqn_to_fsdp_param_info[fqn]
                    if check_all_unflat_param_names_match(optim_state_key.unflat_param_names, fqn2spec_info):
                        unflat_state = _unflatten_optim_state(
                            fsdp_param_info,
                            optim_state_dict[param_key],
                            to_save,
                            False,
                            cpu_offload,
                        )
                    else:
                        unflat_state = _unflatten_optim_state(
                            fsdp_param_info,
                            optim_state_dict[param_key],
                            to_save,
                            shard_state,
                            cpu_offload,
                        )
                    if to_save:
                        assert len(unflat_state) == len(optim_state_key.unflat_param_names)
                        for unflat_param_name, unflat_param_state in zip(
                            optim_state_key.unflat_param_names,
                            unflat_state,
                        ):
                            fsdp_osd_state[unflat_param_name] = unflat_param_state
                elif to_save:
                    assert len(optim_state_key.unflat_param_names) == 1
                    unflat_param_name = optim_state_key.unflat_param_names[0]
                    fsdp_osd_state[unflat_param_name] = copy.copy(optim_state_dict[param_key])
                    if cpu_offload:
                        for state_name, value in sorted_items(fsdp_osd_state[unflat_param_name]):
                            if not torch.is_tensor(value):
                                continue
                            fsdp_osd_state[unflat_param_name][state_name] = value.cpu()

            return fsdp_osd_state

        # monkey patch
        torch.distributed.fsdp._optim_utils._convert_state_with_flat_params = partial(
            _convert_state_with_flat_params_patch, fqn2spec_info=self.fqn2spec_info
        )

    def patch_fsdp_optim_state_dict(self):
        """ """

        def fsdp_optim_state_post_patch_fn(
            model, optim, optim_state_dict=None, fqn2spec_info: Dict[str, SpecInfo] = None
        ):
            assert fqn2spec_info is not None, "if fqn2spec_info is None it should not be patch"

            fsdp_mesh = model._device_mesh
            assert fsdp_mesh is not None, "Please init FSDP module with device_mesh"
            # NOTE we don't support diverse process group for different FSDP sub-modules
            fsdp_pg = model.process_group
            optim_state = orig_optim_state_dict(model, optim, optim_state_dict, fsdp_pg)
            if self.ep_mesh is None:
                return optim_state

            global_device_mesh = self.ep_fsdp_device_mesh
            assert global_device_mesh.ndim == 2

            # extend placements by adding EP placement
            for fqn in sorted(optim_state["state"].keys()):
                if fqn in fqn2spec_info and isinstance(fqn2spec_info[fqn].placement, Shard):
                    cur_spec_info = fqn2spec_info[fqn]
                    fqn_state = {}
                    for key, val in optim_state["state"][fqn].items():
                        # key in OPTIM_STATE_NO_SHARD_KEY in optim stat dict is scalar, like'step', should not be sharded
                        if key not in OPTIM_STATE_NO_SHARD_KEY:
                            val = _shard_tensor(val, cur_spec_info.ep_fsdp_mesh, cur_spec_info.placement)
                        fqn_state[key] = val
                    optim_state["state"][fqn] = fqn_state
            return optim_state

        # monkey patch
        FSDP.optim_state_dict = staticmethod(partial(fsdp_optim_state_post_patch_fn, fqn2spec_info=self.fqn2spec_info))

    def patch_fsdp_optim_state_dict_to_load(self):
        """
        post optimizer state dict hook when calling `FSDP.optim_state_dict(model, optimizer)`

        This will extend the DTensors in optimizer state dict with EP placements

        Args:
            fsdp_no_shard_param_names: List[str], like
        """

        def optim_state_dict_to_load_pre_patch_fn(
            model,
            optim,
            optim_state_dict,
            is_named_optimizer=False,
            load_directly=False,
            group=None,
            fqn2spec_info: Dict[str, SpecInfo] = None,
        ):
            """
            At this point, the `optim_state_dict` is correctly resharded to the current device mesh by `dcp.load`
            """
            assert fqn2spec_info is not None, "if fqn2spec_info is None it should not be patch"

            fsdp_mesh = model._device_mesh
            assert fsdp_mesh is not None, "Please init FSDP module with device_mesh"

            global_device_mesh = self.ep_fsdp_device_mesh
            assert global_device_mesh.ndim == 2

            # NOTE we don't support diverse process group for different FSDP sub-modules
            if self.ep_mesh is not None and self.ep_mesh.size() == self.ep_fsdp_device_mesh.size():
                for fqn in sorted(optim_state_dict["state"].keys()):
                    if check_any_unflat_param_names_match(fqn, fqn2spec_info):
                        fqn_state = {}
                        for key, val in optim_state_dict["state"][fqn].items():
                            # key in OPTIM_STATE_NO_SHARD_KEY in optim stat dict is scalar, like 'step', should not be sharded
                            if key not in OPTIM_STATE_NO_SHARD_KEY:
                                val = _shard_dtensor(val, self.ep_mesh)
                            fqn_state[key] = val
                        optim_state_dict["state"][fqn] = fqn_state

            fsdp_pg = model.process_group
            optim_state = orig_optim_state_dict_to_load(
                model, optim, optim_state_dict, is_named_optimizer, load_directly, fsdp_pg
            )
            return optim_state

        # monkey patch
        FSDP.optim_state_dict_to_load = staticmethod(
            partial(optim_state_dict_to_load_pre_patch_fn, fqn2spec_info=self.fqn2spec_info)
        )


def register_checkpoint_extension(
    fsdp_model: FSDP,
    save_hook_mesh: DeviceMesh = None,
    fqn2spec_info: Dict[str, SpecInfo] = None,
):
    """
    Register dtensor-based hooks for FSDP+EP

    This will:

    1. Customize the FSDP extension for save / load hooks in EP scenarios.
    """

    extension = CheckpointExtensions(
        ep_fsdp_device_mesh=save_hook_mesh,
        fqn2spec_info=fqn2spec_info,
    )
    for fsdp_module in FSDP.fsdp_modules(fsdp_model):
        fsdp_module._fsdp_extension = extension
        fsdp_module._handle._fsdp_extension = extension
    # make sure the root module is also registered
    fsdp_model._fsdp_extension = extension
    fsdp_model._handle._fsdp_extension = extension

    # register load / save hook for ep
    if fqn2spec_info is not None:
        state_dict_post_hook_fn = partial(extension.state_dict_post_hook, fqn2spec_info=fqn2spec_info)
        fsdp_model._register_state_dict_hook(state_dict_post_hook_fn)

        load_state_dict_pre_hook_fn = partial(extension.load_state_dict_pre_hook, fqn2spec_info=fqn2spec_info)
        fsdp_model._register_load_state_dict_pre_hook(load_state_dict_pre_hook_fn)

        # patch load / save functino for ep
        extension.patch_convert_state_with_flat_params()
        extension.patch_fsdp_optim_state_dict()
        extension.patch_fsdp_optim_state_dict_to_load()

    return fsdp_model



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional, Tuple

import torch
import torch.distributed as dist

from .comm import (
    get_unified_sequence_parallel_group,
    get_unified_sequence_parallel_world_size,
)


class ReduceLoss(torch.autograd.Function):
    @staticmethod
    def forward(ctx: torch.autograd.Function, loss: torch.Tensor, num_valid_tokens: torch.Tensor) -> torch.Tensor:
        if num_valid_tokens == 0:
            loss = torch.nan_to_num(loss)

        local_num_tokens = num_valid_tokens.detach().clone()
        loss *= num_valid_tokens
        group = get_unified_sequence_parallel_group()
        dist.all_reduce(loss, group=group)
        dist.all_reduce(num_valid_tokens, group=group)
        ctx.save_for_backward(local_num_tokens, num_valid_tokens)
        return loss / num_valid_tokens

    @staticmethod
    def backward(
        ctx: torch.autograd.Function, grad_output: torch.Tensor
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        local_num_tokens, global_num_tokens = ctx.saved_tensors
        grad_output = get_unified_sequence_parallel_world_size() * local_num_tokens * grad_output / global_num_tokens
        return grad_output, None


def reduce_sequence_parallel_loss(loss: torch.Tensor, num_valid_tokens: torch.Tensor) -> torch.Tensor:
    return ReduceLoss.apply(loss, num_valid_tokens)



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json
import math
import os
from collections import defaultdict
from typing import Callable, Dict, Union

import torch
import torch.distributed as dist
import torch.nn as nn
from safetensors.torch import load_file
from torch.distributed._tensor import Replicate, Shard

from ...utils import logging
from ..parallel_plan import SpecInfo


logger = logging.get_logger(__name__)


def parallel_load_safetensors(
    filepath: str, specific_param_name: list[str] = None, ignore_param_name: list[str] = None
):
    assert not (specific_param_name is not None and ignore_param_name is not None)

    dist.barrier()

    safetensors2param = {}
    index_file = os.path.join(filepath, "model.safetensors.index.json")
    if os.path.exists(index_file):
        index = json.load(open(index_file, "rb"))
        for param_name, filename in index["weight_map"].items():
            if specific_param_name is not None:
                if param_name not in specific_param_name:
                    continue
            elif ignore_param_name is not None:
                if param_name in ignore_param_name:
                    continue
            safetensors2param.setdefault(filename, []).append(param_name)
    else:
        # in this case, the model is small and we can load it all at once
        param_file = os.path.join(filepath, "model.safetensors")
        assert os.path.exists(param_file), f"Cannot find {param_file}"
        states = load_file(param_file)
        for param_name in states:
            safetensors2param.setdefault("model.safetensors", []).append(param_name)
        del states

    total_files = len(safetensors2param)
    ckpt_chunks = sorted(safetensors2param.keys())
    world_size = dist.get_world_size()
    size = int(math.ceil(total_files / world_size))
    ckpt_chunks = [ckpt_chunks[i * size : (i + 1) * size] for i in range(world_size)]

    shard_states = {}
    device = torch.cuda.current_device()
    for rank, files in enumerate(ckpt_chunks):
        if rank == dist.get_rank():
            for file in files:
                safetensors_file = os.path.join(filepath, file)
                states = load_file(safetensors_file, device=device)
                valid_states = {k: v for k, v in states.items() if k in safetensors2param[file]}
                shard_states.update(valid_states)
                del states
        else:
            for file in files:
                for param_name in safetensors2param[file]:
                    shard_states[param_name] = rank
    return shard_states


def parallel_init_fsdp_fn(
    module: torch.nn.Module,
    shard_states: Dict[str, torch.nn.Parameter],
    remove_standalone: bool = True,
    specific_param_name: list[str] = None,
    ignore_param_name: list[str] = None,
):
    """
    Initialize a module with sharded states in a parallel fashion using Fully Sharded Data Parallel (FSDP).

    Args:
        module (torch.nn.Module): The module to be initialized.
        shard_states (Dict[str, torch.nn.Parameter]): A dictionary containing sharded states.
        remove_standalone (bool, optional): If True, only consider shared states. Defaults to True.
        specific_param_name (list[str], optional): A list of specific parameter names to consider. Defaults to None.
        ignore_param_name (list[str], optional): A list of parameter names to ignore. Defaults to None.

    Returns:
        Callable[[torch.nn.Module], torch.nn.Module]: A function that initializes sub-modules of the given module.
    """
    assert not (specific_param_name is not None and ignore_param_name is not None)
    state2fqn = {}
    for name, state in itertools.chain(
        module.named_parameters(remove_duplicate=False), module.named_buffers(remove_duplicate=False)
    ):
        if specific_param_name is not None:
            if name not in specific_param_name:
                continue
        elif ignore_param_name is not None:
            if name in ignore_param_name:
                continue
        state2fqn.setdefault(state, []).append(name)

    shared = {s for s, names in state2fqn.items() if len(names) > 1} if remove_standalone else set(state2fqn.keys())

    materialized_states = {}

    def make_full_tensor(param: torch.Tensor, spec_info: SpecInfo):
        """
        Create a full tensor from a sharded tensor based on the given specification.

        Args:
            param (torch.Tensor): The sharded tensor.
            spec_info (SpecInfo): The specification information.

        Returns:
            torch.Tensor: The full tensor.
        """
        device = torch.cuda.current_device()
        if isinstance(spec_info.placement, Replicate):
            return torch.empty_like(param.data, device=device)
        else:
            assert isinstance(spec_info.placement, Shard)
            size = list(param.shape)
            size[spec_info.placement.dim] *= spec_info.ep_mesh.size()
            return torch.empty(size, dtype=param.dtype, device=device)

    def copy_to_local(param: torch.Tensor, full_data: torch.Tensor, spec_info: SpecInfo):
        """
        Copy data from a full tensor to a local sharded tensor based on the given specification.

        Args:
            param (torch.Tensor): The local sharded tensor.
            full_data (torch.Tensor): The full tensor.
            spec_info (SpecInfo): The specification information.
        """
        if isinstance(spec_info.placement, Replicate):
            param.data.copy_(full_data)
        else:
            assert isinstance(spec_info.placement, Shard)
            local_data = full_data.chunk(spec_info.ep_mesh.size(), dim=spec_info.placement.dim)[
                spec_info.ep_mesh.get_local_rank()
            ]
            param.data.copy_(local_data.contiguous())
        param.spec_info = spec_info

    @torch.no_grad()
    def create_and_sync_state(param_name, state, is_param):
        """
        Create and synchronize a state tensor across multiple devices.

        Args:
            param_name (str): The name of the parameter.
            state (torch.Tensor): The state tensor.
            is_param (bool): Whether the state is a parameter or a buffer.

        Returns:
            torch.Tensor: The synchronized state tensor.
        """
        device = torch.cuda.current_device()
        if is_param:
            param = torch.nn.Parameter(torch.empty_like(state.data, device=device), requires_grad=state.requires_grad)
        else:  # buffer
            param = torch.empty_like(state.data, device=device)
        if param_name not in shard_states:
            logger.warn(f"{param_name} not found in shard states, init it from random")
            assert is_param
            if dist.get_rank() == 0:
                initializer_range = (2.5 * max(state.shape)) ** -0.5
                size = list(state.size())
                if hasattr(state, "spec_info"):
                    shard = state.spec_info.placement
                    if isinstance(shard, Shard):
                        size[shard.dim] *= state.spec_info.ep_mesh.size()
                shard_states[param_name] = torch.nn.Parameter(
                    torch.randn(size, dtype=state.dtype, device=device, requires_grad=state.requires_grad)
                    * initializer_range
                )
            else:
                shard_states[param_name] = 0
        loaded = shard_states[param_name]
        if isinstance(loaded, (torch.nn.Parameter, torch.Tensor)):
            loaded = loaded.to(dtype=param.dtype, device=device)
            dist.broadcast(loaded.data.to(param.dtype), src=dist.get_rank())
            if hasattr(state, "spec_info"):
                copy_to_local(param, loaded.data, state.spec_info)
            else:
                param.data.copy_(loaded.data)
        else:
            assert isinstance(loaded, int)  # the rank that holds the state
            if hasattr(state, "spec_info"):
                full_data = make_full_tensor(param, state.spec_info)
                dist.broadcast(full_data, src=loaded)
                copy_to_local(param, full_data, state.spec_info)
            else:
                dist.broadcast(param.data, src=loaded)
        shard_states.pop(param_name)
        del loaded
        return param

    def init_fn(sub_mod: torch.nn.Module):
        """
        Initialize a sub-module with sharded states.

        Args:
            sub_mod (torch.nn.Module): The sub-module to be initialized.

        Returns:
            torch.nn.Module: The initialized sub-module.
        """
        param_and_buffers = tuple(sub_mod.named_parameters(recurse=False)) + tuple(
            sub_mod.named_buffers(recurse=False)
        )
        for name, state in param_and_buffers:
            if state not in state2fqn:
                logger.warning_once(f"{name} in {sub_mod.__class__.__name__} not found in state2fqn, skip it")
                continue
            is_param = name in sub_mod._parameters
            fqn = state2fqn[state].pop(0)
            if (not is_param) and fqn not in shard_states:
                if state.is_meta:
                    raise RuntimeError(
                        f"find a non-persistent buffer ({fqn}) initiated with device meta. "
                        "Such buffer is not saved in checkpoint and user should guarantee to init in CPU / GPU device."
                    )
                continue
            if state in shared:
                if state not in materialized_states:
                    materialized_states[state] = create_and_sync_state(fqn, state, is_param)
                else:
                    if fqn in shard_states:
                        shard_states.pop(fqn)
                materialize_state = materialized_states[state]
            else:
                materialize_state = create_and_sync_state(fqn, state, is_param)
            if is_param:
                sub_mod._parameters[name] = materialize_state
            else:
                sub_mod._buffers[name] = materialize_state
        return sub_mod

    return init_fn


def init_fsdp_fn(model: nn.Module, device: Union[str, "torch.device"]) -> Callable[[nn.Module], None]:
    """
    Gets tensor materialization function that supports shared parameters and buffers.
    Args:
        model (nn.Module): the top module that may include shared parameters / buffers.
        device (Union[str, torch.device]): the device to initialize parameters on.

    Returns:
        Callable[[nn.Module], None]: initialization method to materialize meta tensors on device.
    """
    param_occurrence = defaultdict(int)
    for _, param in model.named_parameters(remove_duplicate=False):
        param_occurrence[param] += 1

    duplicated_params = {param for param in param_occurrence.keys() if param_occurrence[param] > 1}
    materialized_params = {}

    def init_fn(module: "nn.Module"):
        for name, param in module.named_parameters(recurse=False):
            if param in duplicated_params:
                module._parameters[name] = materialized_params.setdefault(
                    param, nn.Parameter(torch.empty_like(param.data, device=device), requires_grad=param.requires_grad)
                )
            else:
                module._parameters[name] = nn.Parameter(
                    torch.empty_like(param.data, device=device), requires_grad=param.requires_grad
                )

    return init_fn



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, List, Sequence

import torch

from ..utils import logging
from .constants import IGNORE_INDEX


if TYPE_CHECKING:
    from transformers import PreTrainedTokenizer


logger = logging.get_logger(__name__)

ROLE_SUPPORTED = ["system", "user", "assistant", "tool"]


class ChatTemplate(ABC):
    """
    Abstract class for chat template.
    """

    def __init__(self, tokenizer: "PreTrainedTokenizer") -> None:
        self.tokenizer = tokenizer

    def save_pretrained(self, output_dir: str) -> None:
        self.tokenizer.chat_template = self.get_jinja_template()
        try:
            self.tokenizer.save_pretrained(output_dir)
        except Exception:
            logger.warning("Failed to save tokenizer.")

    @abstractmethod
    def encode_messages(self, messages: Sequence[Dict[str, str]], max_seq_len: int = 8192) -> Dict[str, List[int]]:
        """
        Encodes messages to a dictionary of input_ids, attention_mask, and labels.
        """
        ...

    @abstractmethod
    def get_jinja_template(self) -> str:
        """
        Gets the jinja template for the chat template.
        """
        ...


class DefaultTemplate(ChatTemplate):
    def encode_messages(self, messages: Sequence[Dict[str, str]], max_seq_len: int = 8192) -> Dict[str, List[int]]:
        input_ids, attention_mask, labels = [], [], []
        for message in messages:
            content_str = message["role"].title() + ": " + message["content"].strip() + self.tokenizer.eos_token + "\n"
            content_ids = self.tokenizer.encode(content_str, add_special_tokens=False)
            input_ids += content_ids
            attention_mask += [1] * len(content_ids)
            if message["loss_mask"] == 1:
                labels += content_ids
            else:
                labels += [IGNORE_INDEX] * len(content_ids)

        model_inputs = {"input_ids": input_ids, "attention_mask": attention_mask, "labels": labels}
        model_inputs = {k: v[-max_seq_len:] for k, v in model_inputs.items()}
        return model_inputs

    def get_jinja_template(self) -> str:
        return (
            "{% if not add_generation_prompt is defined %}{% set add_generation_prompt = false %}{% endif %}"
            "{% for message in messages %}"
            "{{ message['role'].title() + ': ' + message['content'] | trim + eos_token + '\n' }}"
            "{% endfor %}"
            "{% if add_generation_prompt %}{{ 'Assistant: ' }}{% endif %}"
        )


class Llama2Template(ChatTemplate):
    def encode_messages(self, messages: Sequence[Dict[str, str]], max_seq_len: int = 8192) -> Dict[str, List[int]]:
        input_ids, attention_mask, labels = [], [], []
        for message in messages:
            if message["role"] == "system":
                content_str = "<<SYS>>\n" + message["content"].strip() + "\n<</SYS>>\n\n"
            elif message["role"] == "user":
                content_str = self.tokenizer.bos_token + "[INST] " + message["content"].strip() + " [/INST]"
            elif message["role"] == "assistant":
                content_str = " " + message["content"].strip() + " " + self.tokenizer.eos_token
            elif message["role"] == "tool":
                content_str = self.tokenizer.bos_token + "[TOOL] " + message["content"].strip() + " [/TOOL]"
            else:
                raise ValueError(
                    f"Unknown role {message['role']}, should be one of {{system, user, assistant, tool}}."
                )

            content_ids = self.tokenizer.encode(content_str, add_special_tokens=False)
            input_ids += content_ids
            attention_mask += [1] * len(content_ids)
            if message["loss_mask"] == 1:
                labels += content_ids
            else:
                labels += [IGNORE_INDEX] * len(content_ids)

        model_inputs = {"input_ids": input_ids, "attention_mask": attention_mask, "labels": labels}
        model_inputs = {k: v[-max_seq_len:] for k, v in model_inputs.items()}
        return model_inputs

    def get_jinja_template(self) -> str:
        return (
            "{% if messages[0]['role'] == 'system' %}"
            "{{ '<<SYS>>\n' + messages[0]['content'] | trim + '\n<</SYS>>\n\n' }}"
            "{% set loop_messages = messages[1:] %}"
            "{% else %}"
            "{% set loop_messages = messages %}"
            "{% endif %}"
            "{% for message in loop_messages %}"
            "{% set content = message['content'] %}"
            "{% if message['role'] == 'user' %}"
            "{{ bos_token + '[INST] ' + content | trim + ' [/INST]' }}"
            "{% elif message['role'] == 'tool' %}"
            "{{ bos_token + '[TOOL] ' + content | trim + ' [/TOOL]' }}"
            "{% elif message['role'] == 'assistant' %}"
            "{{ ' ' + content | trim + ' ' + eos_token }}"
            "{% endif %}"
            "{% endfor %}"
        )


class JanusTemplate(ChatTemplate):
    def encode_messages(
        self, messages: Sequence[Dict[str, str]], max_seq_len: int = 8192, task_type: str = ""
    ) -> Dict[str, List[int]]:
        input_ids, attention_mask, labels = [], [], []
        images_seq_mask, images_emb_mask = [], []
        seps = ["\n\n", "<｜end▁of▁sentence｜>"]
        assitant_cnt = 0
        for idx, message in enumerate(messages):
            if message["content"] == "":
                content_str = message["role"] + ":"
            elif (
                "assistant" in message["role"]
                and "wikihow_generation" in task_type
                or "assistant" in message["role"]
                and "interleave_generation" in task_type
            ):
                prefix = "Assistant: " if assitant_cnt == 0 else ""
                suffix = seps[1] if idx + 1 == len(messages) else seps[0]
                content_str = prefix + message["content"].strip() + suffix
                assitant_cnt += 1
            elif "assistant" in message["role"]:
                content_str = "Assistant" + ": " + message["content"].strip() + seps[1]
            elif "user" in message["role"]:
                content_str = "User" + ": " + message["content"].strip() + seps[0]
            elif "system" in message["role"] and "wikihow_generation" in task_type:
                content_str = (
                    message["content"].strip()
                    + seps[0]
                    + "Please generate a step-by-step tutorial with images for the following question."
                    + seps[0]
                )
            elif "system" in message["role"]:
                content_str = message["content"].strip() + seps[0]
            if "system" in message["role"]:
                content_ids = self.tokenizer.encode(content_str)
            else:
                content_ids = self.tokenizer.encode(content_str, add_special_tokens=False)
            input_ids += content_ids
            attention_mask += [1] * len(content_ids)
            image_token_id = self.tokenizer.vocab.get("<image_placeholder>")
            content_ids_tensor = torch.tensor(content_ids)
            images_seq_mask += (content_ids_tensor == image_token_id).tolist()
            image_token_id = self.tokenizer.vocab.get("<image_placeholder>")
            num_image_tokens = torch.sum(content_ids_tensor == image_token_id).item()
            n_image = num_image_tokens // 576
            if n_image > 0:
                for j, n_image_tokens in enumerate([num_image_tokens]):
                    images_emb_mask.append([True] * n_image_tokens)

            if message["loss_mask"] == 1:
                if (
                    image_token_id in content_ids
                    and "wikihow_generation" not in task_type
                    and "interleave_generation" not in task_type
                ):
                    labels += [image_token_id if x == image_token_id else IGNORE_INDEX for x in content_ids]
                else:
                    labels += content_ids
            else:
                labels += [IGNORE_INDEX] * len(content_ids)

        model_inputs = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "labels": labels,
            "images_seq_mask": images_seq_mask,
            "images_emb_mask": images_emb_mask,
        }
        model_inputs = {k: v[-max_seq_len:] for k, v in model_inputs.items()}
        return model_inputs

    def get_jinja_template(self) -> str:
        return (
            "{% if not add_generation_prompt is defined %}{% set add_generation_prompt = false %}{% endif %}"
            "{% for message in messages %}"
            "{{ '<|im_start|>' + message['role'] + '\n' + message['content'] | trim + '<|im_end|>\n' }}"
            "{% endfor %}"
            "{% if add_generation_prompt %}{{ '<|im_start|>assistant\n' }}{% endif %}"
        )


class ChatmlTemplate(ChatTemplate):
    def encode_messages(self, messages: Sequence[Dict[str, str]], max_seq_len: int = 8192) -> Dict[str, List[int]]:
        input_ids, attention_mask, labels = [], [], []
        for message in messages:
            content_str = "<|im_start|>" + message["role"] + "\n" + message["content"].strip() + "<|im_end|>\n"
            content_ids = self.tokenizer.encode(content_str, add_special_tokens=False)
            input_ids += content_ids
            attention_mask += [1] * len(content_ids)
            if message["loss_mask"] == 1:
                labels += content_ids
            else:
                labels += [IGNORE_INDEX] * len(content_ids)

        model_inputs = {"input_ids": input_ids, "attention_mask": attention_mask, "labels": labels}
        model_inputs = {k: v[-max_seq_len:] for k, v in model_inputs.items()}
        return model_inputs

    def get_jinja_template(self) -> str:
        return (
            "{% if not add_generation_prompt is defined %}{% set add_generation_prompt = false %}{% endif %}"
            "{% for message in messages %}"
            "{{ '<|im_start|>' + message['role'] + '\n' + message['content'] | trim + '<|im_end|>\n' }}"
            "{% endfor %}"
            "{% if add_generation_prompt %}{{ '<|im_start|>assistant\n' }}{% endif %}"
        )


TEMPLATES = {
    "default": DefaultTemplate,
    "llama2": Llama2Template,
    "chatml": ChatmlTemplate,
    "Janus": JanusTemplate,
}


def build_chat_template(template_name: str, tokenizer: "PreTrainedTokenizer") -> "ChatTemplate":
    if template_name not in TEMPLATES:
        raise ValueError(f"Unknown chat template: {template_name}")

    return TEMPLATES[template_name](tokenizer)



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from PIL import Image
from transformers import BatchFeature
from transformers.image_utils import ImageInput

from ....transformers.movqgan.processing_movqgan import MoVQGANProcessor
from ..base import BaseDecoderProcessorMixin


class MoVQGANDecoderProcessor(BaseDecoderProcessorMixin, MoVQGANProcessor):
    valid_kwargs = BaseDecoderProcessorMixin.valid_kwargs + ["image_size"]

    def __init__(self, token_size=[1, 32, 32], token_num=1024, image_size=256, **kwargs):
        super().__init__(token_size=token_size, token_num=token_num, image_size=image_size, **kwargs)

    def process(
        self,
        images: Optional[ImageInput] = None,
        return_tensors: str = "pt",
    ) -> BatchFeature:
        features = super().__call__(images=images, return_tensors=return_tensors)["features"]
        image_grid_thw = torch.tensor([self.token_size] * len(images)).type(torch.int32)
        num_image_tokens = [self.token_num] * len(images)
        return BatchFeature(
            data={"features": features, "num_tokens": num_image_tokens, "grid_thw": image_grid_thw},
            tensor_type=return_tensors,
        )

    def postprocess(
        self,
        images: Optional[ImageInput] = None,
        return_tensors: str = "Image",
    ):
        return_image_list = []
        for image in images:
            post_image = image.permute(1, 2, 0)
            post_image = (post_image + 1) * 127.5
            post_image = post_image.to(dtype=torch.uint8)
            if return_tensors == "Image":
                post_image = post_image.detach().cpu().numpy()
                post_image = Image.fromarray(post_image)
            return_image_list.append(post_image)
        if return_tensors != "Image":
            return_image_list = torch.stack(return_image_list, dim=0)
        return return_image_list



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import TYPE_CHECKING, Any, Dict, Literal, Optional

import torch
from transformers import (
    AutoConfig,
    AutoProcessor,
    AutoTokenizer,
    PreTrainedModel,
)

from ..distributed.parallel_state import get_parallel_state
from ..utils import logging
from .loader import BaseModelLoader, get_loader


if TYPE_CHECKING:
    from transformers import PreTrainedTokenizer, ProcessorMixin

logger = logging.get_logger(__name__)


def build_tokenizer(tokenizer_path: str) -> "PreTrainedTokenizer":
    """
    Builds the tokenizer.
    """
    return AutoTokenizer.from_pretrained(tokenizer_path, padding_side="right", trust_remote_code=True)


def build_processor(processor_path: str) -> "ProcessorMixin":
    """
    Builds the processor.
    """
    return AutoProcessor.from_pretrained(processor_path, padding_side="right", trust_remote_code=True)


def build_foundation_model(
    config_path: str,
    weights_path: Optional[str] = None,
    torch_dtype: Literal["float16", "bfloat16", "float32"] = "bfloat16",
    attn_implementation: Optional[Literal["eager", "sdpa", "flash_attention_2"]] = "flash_attention_2",
    moe_implementation: Optional[Literal["eager", "fused"]] = None,
    init_device: Literal["cpu", "cuda", "meta"] = "cuda",
    config_kwargs: Optional[Dict[str, Any]] = None,
    force_use_huggingface: bool = False,
) -> "PreTrainedModel":
    """
    Builds the foundation model.

    If weights_path is provided, it loads the pre-trained weights, otherwise it initializes weights.
    """
    if config_kwargs is None:
        config_kwargs = {}

    config = AutoConfig.from_pretrained(config_path, trust_remote_code=True, **config_kwargs)

    if moe_implementation is not None:
        if moe_implementation not in ["eager", "fused"]:
            raise ValueError(f"Invalid moe_implementation: {moe_implementation}")
        config._moe_implementation = moe_implementation
        logger.info_rank0(f"Moe implementation: {moe_implementation}")

    loader: Optional[BaseModelLoader] = get_loader(config, force_use_huggingface)

    init_kwargs = {
        "config": config,
        "torch_dtype": getattr(torch, torch_dtype),
        "attn_implementation": attn_implementation,
        "trust_remote_code": True,
    }

    if (init_device == "cpu" and get_parallel_state().global_rank != 0) or init_device == "meta":
        empty_init = True
    else:
        empty_init = False

    model = loader.load_model(
        init_kwargs=init_kwargs,
        weights_path=weights_path,
        empty_init=empty_init,
        init_device=init_device,
    )
    return model



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from transformers import BatchFeature, ProcessorMixin

from ...utils import logging


logger = logging.get_logger(__name__)


class SeedOmniProcessor(ProcessorMixin):
    attributes = ["tokenizer"]
    optional_attributes = (
        ["input_image_processor", "input_video_processor", "input_audio_processor"]
        + ["output_image_processor", "output_video_processor", "output_audio_processor"]
        + ["target_image_processor", "target_video_processor", "target_audio_processor"]
        + ["chat_template"]
    )
    input_image_processor_class = "AutoProcessor"
    input_audio_processor_class = "AutoProcessor"
    input_video_processor_class = "AutoProcessor"
    output_image_processor_class = "AutoProcessor"
    output_audio_processor_class = "AutoProcessor"
    output_video_processor_class = "AutoProcessor"
    target_image_processor_class = "AutoProcessor"
    target_audio_processor_class = "AutoProcessor"
    target_video_processor_class = "AutoProcessor"
    tokenizer_class = "AutoTokenizer"
    processor_prefixes = [
        "input_image",
        "output_image",
        "target_image",
        "input_video",
        "output_video",
        "target_video",
        "input_audio",
        "output_audio",
        "target_audio",
    ]

    def __init__(self, tokenizer=None, chat_template=None, **kwargs):
        super().__init__(tokenizer=tokenizer, chat_template=chat_template, **kwargs)

    def __call__(
        self,
        input_images=None,
        output_images=None,
        input_videos=None,
        output_videos=None,
        input_audios=None,
        output_audios=None,
        **kwargs,
    ) -> BatchFeature:
        """For input & output processor:
        features, grid_thw, num_tokens
        For target processor:
        features, kwargs
        """
        inputs = {}
        if input_images:
            image_inputs: BatchFeature = self.input_image_processor.process(images=input_images, **kwargs)
            for key, value in image_inputs.items():
                inputs[f"image_input_{key}"] = value

        if output_images:
            image_inputs: BatchFeature = self.output_image_processor.process(images=output_images, **kwargs)
            for key, value in image_inputs.items():
                inputs[f"image_output_{key}"] = value

            if getattr(self, "target_image_processor", None) is not None:  # target for diffusion models
                image_inputs: BatchFeature = self.target_image_processor.process(images=output_images, **kwargs)
                for key, value in image_inputs.items():
                    inputs[f"image_target_{key}"] = value

        if input_videos or output_videos or input_audios or output_audios:
            raise NotImplementedError

        return BatchFeature(data=inputs)

    def save_pretrained(self, save_directory: str, **kwargs):
        os.makedirs(save_directory, exist_ok=True)
        for prefix in self.processor_prefixes:
            processor = getattr(self, f"{prefix}_processor", None)
            if processor is not None:
                save_path = os.path.join(save_directory, f"{prefix}_processor")
                processor.save_pretrained(save_path)

        return super().save_pretrained(save_directory, **kwargs)

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str, **kwargs):
        processor = super().from_pretrained(pretrained_model_name_or_path, **kwargs)
        # if return_unused_kwargs a tuple is returned where the second element is "unused_kwargs"
        if isinstance(processor, tuple):
            processor = processor[0]

        for prefix in cls.processor_prefixes:
            setattr(processor, f"{prefix}_processor", None)

        return processor



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
import triton
import triton.language as tl

from .triton_utils.memory import (
    load_with_pred_1d,
    store_with_pred_1d,
)


@triton.heuristics(values={"BLOCK_ALIGNED": lambda args: args["num_elts"] % args["BLOCK_SIZE"] == 0})
@triton.jit
def _expert_histogram_kernel(
    out_ptr,
    x_ptr,
    num_elts,
    num_bins,
    NUM_BINS_LAST_UNUSED: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    BLOCK_ALIGNED: tl.constexpr,
):
    pid = tl.program_id(0)

    in_off = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    data = load_with_pred_1d(x_ptr + in_off, BLOCK_ALIGNED, in_off < num_elts, NUM_BINS_LAST_UNUSED - 1).to(tl.int32)

    tl.device_assert(
        data < num_bins or data == NUM_BINS_LAST_UNUSED - 1,
        "Out-of-bound element found.",
    )
    count = tl.histogram(data, NUM_BINS_LAST_UNUSED)

    out_off = tl.arange(0, NUM_BINS_LAST_UNUSED)
    tl.atomic_add(out_ptr + out_off, count, mask=out_off < num_bins, sem="relaxed")


def expert_histogram(input: torch.Tensor, num_bins: int) -> torch.Tensor:
    """Returns histogram of `input`, with bin width 1. Note that for each individual `num_bins`,
    a separate Triton kernel is generated (mostly). So if `num_bins` varies between calls, you
    probably should go for some other histogram method.
    """

    assert input.is_cuda
    assert input.dtype == torch.int32 or input.dtype == torch.int64
    assert input.numel() < (1 << 31) - 1, "Too many elements."
    flattened = input.flatten().contiguous()

    # An extra slot is needed, our kernel uses the extra slot to handle possible OoO reads.
    # Wastes a lot of slots but hopefully the kernel can still saturate memory B/W.
    NUM_BINS_LAST_UNUSED = triton.next_power_of_2(num_bins + 1)
    out = torch.zeros([num_bins], dtype=torch.int32, device=input.device)

    BLOCK_SIZE = 1024
    num_elts = flattened.numel()
    grid = (triton.cdiv(num_elts, BLOCK_SIZE),)
    with torch.cuda.device(input.device):
        _expert_histogram_kernel[grid](
            out_ptr=out,
            x_ptr=flattened,
            num_elts=num_elts,
            num_bins=num_bins,
            NUM_BINS_LAST_UNUSED=NUM_BINS_LAST_UNUSED,
            BLOCK_SIZE=BLOCK_SIZE,
        )

    return out[:num_bins]


@triton.heuristics(values={"N_ALIGNED": lambda args: args["N"] % args["BLOCK_N"] == 0})
@triton.jit
def _moe_gather_kernel(
    X,
    Y,
    index,
    num_elts_in,
    num_elts_out,
    N: tl.constexpr,  # hidden size
    TOPK: tl.constexpr,
    STRIDE_XM: tl.constexpr,
    STRIDE_XN: tl.constexpr,
    STRIDE_OM: tl.constexpr,
    STRIDE_ON: tl.constexpr,
    STRIDE_IM: tl.constexpr,
    STRIDE_IN: tl.constexpr,
    BLOCK_N: tl.constexpr,
    N_ALIGNED: tl.constexpr,
):
    r"""
    X: m * topk x n
    Y: m x n
    index: m x topk
    code:
        repeated-X: m * topk x n -> reduce(sum_over_topk) -> m x n
        Y: Y[arange(m)] = sum_over_topk(repeated-X[arange(m) * topk])
    """
    pid_m = tl.program_id(axis=0).to(tl.int64)  # m
    block_idx = tl.program_id(axis=1).to(tl.int64)
    n = block_idx * BLOCK_N + tl.arange(0, BLOCK_N)
    y = tl.zeros([BLOCK_N], dtype=tl.float32)
    for i in tl.static_range(TOPK):
        x_index = tl.load(index + pid_m.to(tl.int64) * STRIDE_IM + i * STRIDE_IN)
        tl.device_assert(x_index < num_elts_in, "Input OOB")
        x = load_with_pred_1d(
            X + x_index.to(tl.int64) * STRIDE_XM + n.to(tl.int64) * STRIDE_XN, N_ALIGNED, mask=n < N, other=0
        )
        y += x
    # save one line
    tl.device_assert(pid_m < num_elts_out, "Output OOB")
    Y = Y + pid_m.to(tl.int64) * STRIDE_OM + n.to(tl.int64) * STRIDE_ON  # noqa
    store_with_pred_1d(Y, y, N_ALIGNED, mask=n < N)


def moe_gather(x: torch.Tensor, index: torch.Tensor, out_dtype=None):
    assert x.is_cuda and index.is_cuda
    M, topk = index.shape
    assert x.shape[0] == M * topk
    N = x.shape[1]

    assert x.device == index.device, f"x.device = {x.device}, index.device = {index.device}"

    out_dtype = out_dtype or x.dtype
    out = torch.empty(M, N, dtype=out_dtype, device=x.device)

    grid = lambda meta: (M, triton.cdiv(N, meta["BLOCK_N"]))  # noqa
    with torch.cuda.device(x.device):
        _moe_gather_kernel[grid](
            x,
            out,
            index,
            num_elts_in=M * topk,
            num_elts_out=M,
            N=N,
            TOPK=topk,
            STRIDE_XM=x.stride(0),
            STRIDE_XN=x.stride(1),
            STRIDE_OM=out.stride(0),
            STRIDE_ON=out.stride(1),
            STRIDE_IM=index.stride(0),
            STRIDE_IN=index.stride(1),
            BLOCK_N=1024,
        )

    return out


@triton.heuristics(values={"N_ALIGNED": lambda args: args["N"] % args["BLOCK_N"] == 0})
@triton.jit
def _moe_add_gather_kernel(
    X,
    Y,
    Z,
    index,
    num_elts_in,
    num_elts_out,
    N: tl.constexpr,  # hidden size
    TOPK: tl.constexpr,
    STRIDE_XM: tl.constexpr,
    STRIDE_XN: tl.constexpr,
    STRIDE_YM: tl.constexpr,
    STRIDE_YN: tl.constexpr,
    STRIDE_OM: tl.constexpr,
    STRIDE_ON: tl.constexpr,
    STRIDE_IM: tl.constexpr,
    STRIDE_IN: tl.constexpr,
    BLOCK_N: tl.constexpr,
    N_ALIGNED: tl.constexpr,
):
    r"""
    X: m * topk x n
    Y: m * topk x n
    Z: m x n
    index: m x topk

    code:
        repeated-(X + Y): m * topk x n -> reduce(sum_over_topk) -> m x n
        Z: Z[arange(m)] = sum_over_topk(repeated-(X+Y)[arange(m) * topk])
    """
    pid_m = tl.program_id(axis=0)  # m
    block_idx = tl.program_id(axis=1)

    n = block_idx * BLOCK_N + tl.arange(0, BLOCK_N)
    z = tl.zeros([BLOCK_N], dtype=tl.float32)
    for i in tl.static_range(TOPK):
        x_index = tl.load(index + pid_m * STRIDE_IM + i * STRIDE_IN)
        tl.device_assert(x_index < num_elts_in, "Input OOB")
        x = load_with_pred_1d(X + x_index * STRIDE_XM + n * STRIDE_XN, N_ALIGNED, mask=n < N, other=0)
        y = load_with_pred_1d(Y + x_index * STRIDE_YM + n * STRIDE_YN, N_ALIGNED, mask=n < N, other=0)
        z += x + y

    # save one line
    tl.device_assert(pid_m < num_elts_out, "Output OOB")
    Z = Z + pid_m * STRIDE_OM + n * STRIDE_ON  # noqa
    store_with_pred_1d(Z, z, N_ALIGNED, mask=n < N)


def moe_add_gather(x: torch.Tensor, y: torch.Tensor, index: torch.Tensor, out_dtype=None):
    assert x.is_cuda and y.is_cuda and index.is_cuda
    assert x.shape == y.shape
    assert x.dtype == y.dtype
    M, topk = index.shape
    assert x.shape[0] == M * topk
    N = x.shape[1]

    assert x.device == y.device, f"x.device = {x.device}, y.device = {y.device}"
    assert x.device == index.device, f"x.device = {x.device}, index.device = {index.device}"

    out_dtype = out_dtype or x.dtype
    out = torch.empty(M, N, dtype=out_dtype, device=x.device)

    grid = lambda meta: (M, triton.cdiv(N, meta["BLOCK_N"]))  # noqa
    with torch.cuda.device(x.device):
        _moe_add_gather_kernel[grid](
            x,
            y,
            out,
            index,
            num_elts_in=M * topk,
            num_elts_out=M,
            N=N,
            TOPK=topk,
            STRIDE_XM=x.stride(0),
            STRIDE_XN=x.stride(1),
            STRIDE_YM=y.stride(0),
            STRIDE_YN=y.stride(1),
            STRIDE_OM=out.stride(0),
            STRIDE_ON=out.stride(1),
            STRIDE_IM=index.stride(0),
            STRIDE_IN=index.stride(1),
            BLOCK_N=1024,
        )

    return out


@triton.heuristics(values={"N_ALIGNED": lambda args: args["N"] % args["BLOCK_N"] == 0})
@triton.jit
def _moe_scatter_kernel(
    X,
    O,  # noqa
    index,
    num_elts_in,
    num_elts_out,
    N: tl.constexpr,  # hidden size
    TOPK: tl.constexpr,
    STRIDE_XM: tl.constexpr,
    STRIDE_XN: tl.constexpr,
    STRIDE_OM: tl.constexpr,
    STRIDE_ON: tl.constexpr,
    STRIDE_IM: tl.constexpr,
    STRIDE_IN: tl.constexpr,
    BLOCK_N: tl.constexpr,
    N_ALIGNED: tl.constexpr,
):
    r"""
    X: m x n
    O: m * topk x n
    index: m x topk

    code:
        X: m x n -> repeat -> m x topk x n -> m * topk x n
            X[arange(m) * topk] = X[arange(m)]

        O[index] = X
            O[index[arange(m) * topk]] = X[arange(m) * topk]
    """

    pid_m = tl.program_id(axis=0)  # m
    block_idx = tl.program_id(axis=1)
    n = block_idx * BLOCK_N + tl.arange(0, BLOCK_N)

    tl.device_assert(pid_m < num_elts_in, "Input OOB.")
    X = X + pid_m * STRIDE_XM + n * STRIDE_XN
    x = load_with_pred_1d(X, N_ALIGNED, mask=n < N, other=0)

    for i in tl.static_range(TOPK):
        o_index = tl.load(index + pid_m * STRIDE_IM + i * STRIDE_IN)
        tl.device_assert(o_index < num_elts_out, "Output OOB.")
        tmp_index = o_index.to(tl.int64) * STRIDE_OM
        # tl.device_print("tmp_index", tmp_index)
        out = O + tmp_index + n * STRIDE_ON

        # save one line
        store_with_pred_1d(out, x, N_ALIGNED, mask=n < N)


def moe_scatter(x: torch.Tensor, index: torch.Tensor, out_dtype=None):
    assert x.is_cuda and index.is_cuda
    assert x.shape[0] == index.shape[0]

    assert x.device == index.device, f"x.device = {x.device}, index.device = {index.device}"

    M, N = x.shape
    topk = index.shape[1]
    out_dtype = out_dtype or x.dtype
    out = torch.empty(M * topk, N, dtype=out_dtype, device=x.device)
    assert lambda: index.unique().numel() == M * topk, "Holes in output?"

    grid = lambda meta: (M, triton.cdiv(N, meta["BLOCK_N"]))  # noqa
    with torch.cuda.device(x.device):
        _moe_scatter_kernel[grid](
            x,
            out,
            index,
            num_elts_in=M,
            num_elts_out=M * topk,
            N=N,
            TOPK=topk,
            STRIDE_XM=x.stride(0),
            STRIDE_XN=x.stride(1),
            STRIDE_OM=out.stride(0),
            STRIDE_ON=out.stride(1),
            STRIDE_IM=index.stride(0),
            STRIDE_IN=index.stride(1),
            BLOCK_N=1024,
        )

    return out


@triton.jit
def _moe_index_compute_kernel(
    indices_ptr,
    experts_for_tokens_ptr,
    temp_histogram_cumsum_ptr,
    num_elts,
    NUM_EXPERTS: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,  # Unlikely to be aligned, so we don't test for alignment.
):
    _OOB_EXPERT_ID: tl.constexpr = 1023
    tl.static_assert(_OOB_EXPERT_ID > NUM_EXPERTS, "Too many experts for me.")

    start_pos = tl.program_id(0)
    processing_range = start_pos * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    expert_ids = tl.load(
        experts_for_tokens_ptr + processing_range,
        processing_range < num_elts,
        _OOB_EXPERT_ID,
    )
    assert expert_ids < NUM_EXPERTS or expert_ids == _OOB_EXPERT_ID

    indices = tl.zeros([BLOCK_SIZE], dtype=tl.int32)
    for expert_id in tl.static_range(NUM_EXPERTS):
        mask = expert_ids == expert_id
        one_if_expert_id_matches = mask.to(tl.int32)

        # Tokens allocated to this expert.
        slots_to_reserve = tl.sum(one_if_expert_id_matches)
        slot_ids = (
            # Reserve last `slots_to_reserve` slots for us.
            tl.atomic_add(temp_histogram_cumsum_ptr + expert_id, -slots_to_reserve, sem="relaxed")
            # `atomic_add` returns old value, so we need to do substraction again.
            - slots_to_reserve
            # Local offset for each token in `expert_ids`.
            + tl.cumsum(one_if_expert_id_matches)
            # Result of `cumsum` is "1-based".
            - 1
        )
        assigned_slot_or_zero = tl.where(mask, slot_ids, 0)
        indices += assigned_slot_or_zero.to(tl.int32)

    tl.store(indices_ptr + processing_range, indices, processing_range < num_elts)


def moe_index_compute(experts_for_tokens: torch.Tensor, expert_histogram_cumsum: torch.Tensor) -> torch.Tensor:
    """Calculate row number into activation passed to MoE fc1 for each token.

    Arguments:

        - experts_for_tokens: [n_tokens, expert_topk] experts assigned to each token.
        - expert_histogram_cumsum: [n_experts]: cumsum of number of tokens allocated to each expert,
          with last element being number of tokens. NOTE: This is usually calculated as part of gemm
          grouped, so you can just reuse it.

    Returns:

        - [n_tokens, expert_topk] row number into activation passed to MoE fc1 for each token. Each
          token should be duplicated `expert_topk` times.
    """
    # No noncontiguous input.
    assert experts_for_tokens.is_contiguous()
    assert experts_for_tokens.numel() < (1 << 31) - 1
    assert expert_histogram_cumsum.is_contiguous()
    assert experts_for_tokens.device == expert_histogram_cumsum.device, (
        f"experts_for_tokens.device = {experts_for_tokens.device}, expert_histogram_cumsum.device = {expert_histogram_cumsum.device}"
    )

    BLOCK_SIZE = 128  # Faster than 1024, not sure why. May be better occupancy?

    histogram_cumsum_copy = expert_histogram_cumsum.clone().detach()  # Temporary workspace.
    indices = torch.empty_like(experts_for_tokens, dtype=int)

    with torch.cuda.device(experts_for_tokens.device):
        _moe_index_compute_kernel[(triton.cdiv(experts_for_tokens.numel(), BLOCK_SIZE),)](
            indices_ptr=indices,
            experts_for_tokens_ptr=experts_for_tokens,
            temp_histogram_cumsum_ptr=histogram_cumsum_copy,
            num_elts=experts_for_tokens.numel(),
            NUM_EXPERTS=histogram_cumsum_copy.numel(),
            BLOCK_SIZE=BLOCK_SIZE,
        )

    return indices



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from typing import Optional

from typing import Optional

import numpy as np
import torch
from transformers import BatchFeature, ProcessorMixin
from transformers.image_transforms import center_crop, resize
from transformers.image_utils import ImageInput, PILImageResampling, make_list_of_images, to_numpy_array


class MoVQGANProcessor(ProcessorMixin):
    attributes = []

    def __init__(self, image_size: int = 256, **kwargs):
        super().__init__(**kwargs)
        self.image_size = image_size

    def transform(self, image):
        crop_size = min(image.shape[0], image.shape[1])
        image = center_crop(image, (crop_size, crop_size), input_data_format="channels_last")
        image = resize(
            image,
            (self.image_size, self.image_size),
            resample=PILImageResampling.BICUBIC,
            reducing_gap=1,
            input_data_format="channels_last",
        )
        image = image.astype(np.float32) / 127.5 - 1.0
        return np.transpose(image, [2, 0, 1])

    def __call__(
        self,
        images: Optional[ImageInput] = None,
        return_tensors: str = "pt",
    ) -> torch.Tensor:
        images = make_list_of_images(images)
        images = [to_numpy_array(image) for image in images]
        pixel_values = [self.transform(image) for image in images]
        return BatchFeature(
            data={"features": pixel_values},
            tensor_type=return_tensors,
        )
